(* Title:  BaseDefinitions/references.sml
   Author: Walther Neuper
   (c) due to copyright terms

References into Know_Store.
*)
signature REFERENCES =
sig
  type T
  val empty: T
  val to_string: T -> string
  val select_input : T -> T -> T

  val complete: Ctree.state -> Ctree.state
  val are_complete: Ctree.state -> bool
end

(**)
structure References(**): REFERENCES(**) =
struct
(**)

type id = References_Def.id;
type T = References_Def.T;
val empty = References_Def.empty;
val to_string = References_Def.to_string;
val select_input = References_Def.select_input;

(* have (thy, pbl, met) _all_ been specified explicitly ? *)
fun are_complete (pt, pos as (p, _)) = 
  if (not o Ctree.is_pblobj o (Ctree.get_obj I pt)) p then
    raise ERROR ("References.is_complete: called by PrfObj at " ^ Pos.pos'2str pos)
  else
    let
      val (dI, pI, mI) = Ctree.get_obj Ctree.g_spec pt p
    in
      dI <> ThyC.id_empty andalso pI <> Problem.id_empty andalso mI <> MethodC.id_empty
    end;

(* complete References and save them to the state *)
fun complete (pt, pos as (p, _)) = 
  let
    val (ospec, spec) = case Ctree.get_obj I pt p of
      Ctree.PblObj {origin = (_, ospec, _), spec, ...} => (ospec, spec)
    | _ => raise ERROR "References.complete: uncovered case get_obj"
    val pt = Ctree.update_spec pt p (select_input ospec spec)
  in
    (pt, pos)
  end

(**)end(**)
